from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, TypeVar

from pydantic import BaseModel, ConstrainedList, Extra, Field, HttpUrl, validator
from pydantic.generics import GenericModel

if TYPE_CHECKING:
    from pydantic.typing import CallableGenerator


def gen_alias(field_name: str) -> str:
    if field_name == "id_":
        return "id"
    return "".join(
        [w if i == 0 else w.capitalize() for i, w in enumerate(field_name.split("_"))]
    )


class Base(BaseModel):
    class Config:
        allow_population_by_field_name = True
        extra = Extra.forbid
        alias_generator = gen_alias


class CustomList(ConstrainedList):
    item_type: Any

    def __init_subclass__(cls) -> None:
        cls.__args__ = (cls.item_type,)

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        for cls_validator in super().__get_validators__():
            yield cls_validator
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> "CustomList":
        return cls(v)


# query params


class PaginationQueryParams(Base):
    cursor: Optional[str] = None
    limit: Optional[int] = Field(None, le=5000)


class TradeQueryParams(Base):
    since: Optional[int]
    limit: Optional[int]


class MarketSummariesQueryParams(PaginationQueryParams):
    key_by: Optional[str] = Field(alias="keyBy")


class OrderBookQueryParams(Base):
    depth: Optional[int]
    span: Optional[float]
    limit: Optional[int]


class OrderBookCalculatorQueryParams(Base):
    # NOTE: this is float, not Decimal, because it will be sent
    amount: float


class OHLCVQueryParams(Base):
    before: Optional[int]
    after: Optional[int]
    periods: Optional[str]

    @validator("periods", pre=True)
    def list_to_str(cls, v: List[int]) -> Optional[str]:
        if not v:
            return None

        if not all(isinstance(item, int) for item in v):
            raise TypeError("all periods must be integers")

        return ",".join(map(str, v))


# path params


class AssetPathParams(Base):
    asset_code: str


class PairPathParams(Base):
    pair: str


class MarketPathParams(Base):
    exchange: str
    pair: str


class ExchangePathParams(Base):
    exchange: str


# models


ResultT = TypeVar("ResultT")


class Cursor(Base):
    last: str
    has_more: bool = Field(alias="hasMore")


class Allowance(Base):
    cost: float
    remaining: float
    upgrade: str


class Response(GenericModel, Generic[ResultT]):
    result: ResultT
    allowance: Allowance


class Route(HttpUrl):
    pass


class PaginatedResponse(Response[ResultT], Generic[ResultT]):
    cursor: Cursor


class ListMember(Base):
    route: Route


class AssetBase(Base):
    id_: int
    symbol: str
    name: str
    fiat: bool


class AssetMember(ListMember, AssetBase):
    pass


class AssetList(CustomList):
    item_type = AssetMember


class MarketBase(Base):
    id_: int
    exchange: str
    pair: str
    active: bool


class PairBase(Base):
    id_: int
    symbol: str
    base: AssetMember
    quote: AssetMember
    futuresContractPeriod: Optional[str] = None


class PairMember(ListMember, PairBase):
    pass


class PairList(CustomList):
    item_type = PairMember


class MarketMember(ListMember, MarketBase):
    pass


class AssetMarketList(Base):
    base: List[MarketMember]
    quote: List[MarketMember]


class Asset(AssetBase):
    markets: AssetMarketList


class MarketList(CustomList):
    item_type = MarketMember


class Pair(PairBase):
    route: Route
    markets: MarketList


class MarketRoutes(Base):
    price: Route
    summary: Route
    orderbook: Route
    trades: Route
    ohlc: Route


class Market(MarketBase):
    routes: MarketRoutes


class Price(Decimal):
    pass


class MarketPrice(Base):
    price: Price


AllPrices = Dict[str, Price]


class Trade(Base):
    id_: str
    timestamp: int
    price: Price
    amount: Decimal

    @classmethod
    def validate(cls, v: List[Any]) -> "Trade":
        return cls(
            id_=v[0],
            timestamp=v[1],
            price=v[2],
            amount=v[3],
        )


class MarketTradeList(CustomList):
    item_type = Trade


class PriceChange(Base):
    percentage: Decimal
    absolute: Decimal


class PriceSummary(Base):
    last: Decimal
    high: Decimal
    low: Decimal
    change: PriceChange


class MarketSummary(Base):
    price: PriceSummary
    volume: Decimal
    volume_quote: Decimal = Field(alias="volumeQuote")


AllSummaries = Dict[str, MarketSummary]


class OrderBookItem(Base):
    price: Price
    amount: Decimal

    @classmethod
    def validate(cls, v: List[Any]) -> "OrderBookItem":
        return cls(
            price=v[0],
            amount=v[1],
        )


class OrderBookArray(CustomList):
    item_type = OrderBookItem


class OrderBook(Base):
    asks: OrderBookArray
    bids: OrderBookArray
    seq_num: int = Field(alias="seqNum")


class LiquidityLevel(Base):
    base: Dict[int, Decimal]
    quote: Dict[int, Decimal]


class OrderBookLiquidity(Base):
    bid: LiquidityLevel
    ask: LiquidityLevel


class QuoteBase(Base):
    avg_price: Price
    avg_delta: Decimal
    avg_delta_bps: Decimal
    reach_price: Price
    reach_delta: Decimal
    reach_delta_bps: Decimal


class QuoteBuy(QuoteBase):
    spend: Decimal


class QuoteSell(QuoteBase):
    receive: Decimal


class OrderBookCalculator(Base):
    buy: QuoteBuy
    sell: QuoteSell


class OHLCV(Base):
    close_time: int
    open_price: Price
    high_price: Price
    low_price: Price
    close_price: Price
    volume: Decimal
    quote_volume: Decimal

    @classmethod
    def validate(cls, v: List[Any]) -> "OHLCV":
        return cls(
            close_time=v[0],
            open_price=v[1],
            high_price=v[2],
            low_price=v[3],
            close_price=v[4],
            volume=v[5],
            quote_volume=v[6],
        )


OHLCVDict = Dict[str, List[OHLCV]]


class ExchangeBase(Base):
    id_: int
    symbol: str
    name: str
    active: bool


class ExchangeMember(ListMember, ExchangeBase):
    pass


class ExchangeList(CustomList):
    item_type = ExchangeMember


class ExchangeRoutes(Base):
    markets: Route


class Exchange(ExchangeBase):
    routes: ExchangeRoutes


class ExchangeMarkets(CustomList):
    item_type = MarketMember



from apiclient import endpoint


@endpoint(base_url="https://api.cryptowat.ch")
class Endpoint:
    list_assets = "/assets"
    asset_detail = "/assets/{assetCode}"
    list_pairs = "/pairs"
    pair_detail = "/pairs/{pair}"
    list_markets = "/markets"
    market_detail = "/markets/{exchange}/{pair}"
    all_market_prices = "/markets/prices"
    market_price = "/markets/{exchange}/{pair}/price"
    list_market_trades = "/markets/{exchange}/{pair}/trades"
    market_summary = "/markets/{exchange}/{pair}/summary"
    all_market_summaries = "/markets/summaries"
    market_orderbook = "/markets/{exchange}/{pair}/orderbook"
    market_orderbook_liquidity = "/markets/{exchange}/{pair}/orderbook/liquidity"
    market_orderbook_calculator = "/markets/{exchange}/{pair}/orderbook/calculator"
    market_ohlc = "/markets/{exchange}/{pair}/ohlc"
    list_exchanges = "/exchanges"
    exchange_detail = "/exchanges/{exchange}"
    exchange_markets = "/markets/{exchange}"



from typing import Optional

from apiclient import APIClient
from apiclient.authentication_methods import HeaderAuthentication, NoAuthentication
from apiclient.response_handlers import JsonResponseHandler
from apiclient_pydantic import serialize_all_methods

from .config import settings
from .endpoints import Endpoint
from .models import (
    AllPrices,
    AllSummaries,
    Asset,
    AssetList,
    AssetPathParams,
    Exchange,
    ExchangeList,
    ExchangeMarkets,
    ExchangePathParams,
    Market,
    MarketList,
    MarketPathParams,
    MarketPrice,
    MarketSummariesQueryParams,
    MarketSummary,
    MarketTradeList,
    OHLCVDict,
    OHLCVQueryParams,
    OrderBook,
    OrderBookCalculator,
    OrderBookCalculatorQueryParams,
    OrderBookLiquidity,
    OrderBookQueryParams,
    PaginatedResponse,
    PaginationQueryParams,
    Pair,
    PairList,
    PairPathParams,
    Response,
    TradeQueryParams,
)

NO_KEY_MESSAGE = """\
You have not set an API Key. Anonymous users are limited to 10 Cryptowatch
Credits worth of API calls per 24-hour period.

See https://docs.cryptowat.ch/rest-api/rate-limit#api-request-pricing-structure
for more information.\
"""


@serialize_all_methods()
class CryptoWatchClient(APIClient):
    def __init__(self, api_key: Optional[str] = None) -> None:
        api_key = api_key or settings.CW_API_KEY
        self._api_key = api_key
        if not api_key:
            print(NO_KEY_MESSAGE)
            authentication_method = NoAuthentication()
        else:
            authentication_method = HeaderAuthentication(
                token=api_key, parameter="X-CW-API-Key", scheme=None
            )

        super().__init__(
            response_handler=JsonResponseHandler,
            authentication_method=authentication_method,
        )

    @property
    def is_authenticated(self) -> bool:
        return self._api_key is not None

    def list_assets(
        self, params: PaginationQueryParams
    ) -> PaginatedResponse[AssetList]:
        return self.get(Endpoint.list_assets, params=params)

    def get_asset(self, path_params: AssetPathParams) -> Response[Asset]:
        return self.get(Endpoint.asset_detail.format(**path_params))

    def list_pairs(self, params: PaginationQueryParams) -> PaginatedResponse[PairList]:
        return self.get(Endpoint.list_pairs, params=params)

    def get_pair(self, path_params: PairPathParams) -> Response[Pair]:
        return self.get(Endpoint.pair_detail.format(**path_params))

    def list_markets(
        self, params: PaginationQueryParams
    ) -> PaginatedResponse[MarketList]:
        return self.get(Endpoint.list_markets, params=params)

    def get_market(self, path_params: MarketPathParams) -> Response[Market]:
        return self.get(Endpoint.market_detail.format(**path_params))

    def get_market_price(self, path_params: MarketPathParams) -> Response[MarketPrice]:
        return self.get(Endpoint.market_price.format(**path_params))

    def get_all_market_prices(
        self, params: PaginationQueryParams
    ) -> PaginatedResponse[AllPrices]:
        return self.get(Endpoint.all_market_prices, params=params)

    def get_market_trades(
        self, path_params: MarketPathParams, params: TradeQueryParams
    ) -> Response[MarketTradeList]:
        return self.get(
            Endpoint.list_market_trades.format(**path_params), params=params
        )

    def get_market_summary(
        self, path_params: MarketPathParams
    ) -> Response[MarketSummary]:
        return self.get(Endpoint.market_summary.format(**path_params))

    def get_all_market_summaries(
        self, params: MarketSummariesQueryParams
    ) -> Response[AllSummaries]:
        return self.get(Endpoint.all_market_summaries, params=params)

    def get_market_order_book(
        self, path_params: MarketPathParams, params: OrderBookQueryParams
    ) -> Response[OrderBook]:
        return self.get(Endpoint.market_orderbook.format(**path_params), params=params)

    def get_market_order_book_liquidity(
        self, path_params: MarketPathParams
    ) -> Response[OrderBookLiquidity]:
        return self.get(Endpoint.market_orderbook_liquidity.format(**path_params))

    def calculate_quote(
        self, path_params: MarketPathParams, params: OrderBookCalculatorQueryParams
    ) -> Response[OrderBookCalculator]:
        return self.get(
            Endpoint.market_orderbook_calculator.format(**path_params), params=params
        )

    def get_ohlcv(
        self, path_params: MarketPathParams, params: OHLCVQueryParams
    ) -> Response[OHLCVDict]:
        return self.get(Endpoint.market_ohlc.format(**path_params), params=params)

    def list_exchanges(self) -> Response[ExchangeList]:
        return self.get(Endpoint.list_exchanges)

    def get_exchange(self, path_params: ExchangePathParams) -> Response[Exchange]:
        return self.get(Endpoint.exchange_detail.format(**path_params))

    def list_exchange_markets(
        self, path_params: ExchangePathParams
    ) -> Response[ExchangeMarkets]:
        return self.get(Endpoint.exchange_markets.format(**path_params))
